import { StateContext, Logger, StatsContext, TransactionContext, EngageDestinationCache, ActionHookType, SubscriptionMetadata } from './index';
import type { RequestOptions } from '../request-client';
import type { JSONLikeObject, JSONObject } from '../json-object';
import { AuthTokens } from './parse-settings';
import type { RequestClient } from '../create-request-client';
import type { ID } from '../segment-event';
import { Features } from '../mapping-kit';
import type { ErrorCodes, MultiStatusErrorReporter } from '../errors';
export declare type Optional<T, K extends keyof T> = Pick<Partial<T>, K> & Omit<T, K>;
export declare type MaybePromise<T> = T | Promise<T>;
export interface Result {
    output?: JSONObject | string | null | undefined;
    error?: JSONObject | null;
    data?: JSONObject | null;
    multistatus?: ResultMultiStatusNode[];
}
export interface DynamicFieldContext {
    selectedArrayIndex?: number;
    selectedKey?: string;
}
export interface ExecuteInput<Settings, Payload, AudienceSettings = unknown, ActionHookInputs = any, ActionHookOutputs = any> {
    readonly mapping?: JSONObject;
    readonly settings: Settings;
    readonly audienceSettings?: AudienceSettings;
    payload: Payload;
    hookInputs?: ActionHookInputs;
    hookOutputs?: Partial<Record<ActionHookType, ActionHookOutputs>>;
    dynamicFieldContext?: DynamicFieldContext;
    page?: string;
    syncMode?: SyncMode;
    matchingKey?: string;
    readonly auth?: AuthTokens;
    readonly features?: Features;
    readonly statsContext?: StatsContext;
    readonly logger?: Logger;
    readonly engageDestinationCache?: EngageDestinationCache;
    readonly transactionContext?: TransactionContext;
    readonly stateContext?: StateContext;
    readonly subscriptionMetadata?: SubscriptionMetadata;
}
export interface DynamicFieldResponse {
    choices: DynamicFieldItem[];
    nextPage?: string;
    error?: DynamicFieldError;
}
export interface DynamicFieldError {
    code: string;
    message: string;
}
export interface DynamicFieldItem {
    label: string;
    value: string;
    description?: string;
    type?: FieldTypeName;
}
export interface GlobalSetting {
    label: string;
    description: string;
    type: 'boolean' | 'string' | 'password' | 'number';
    multiple?: boolean;
    choices?: Array<{
        value: string | number;
        label: string;
    }>;
    required?: boolean;
    default?: string | number | boolean;
    properties?: InputField['properties'];
    format?: InputField['format'];
    depends_on?: InputField['depends_on'];
}
export declare type FieldTypeName = 'string' | 'text' | 'number' | 'integer' | 'datetime' | 'boolean' | 'password' | 'object';
declare type FieldCategory = 'identifier' | 'data' | 'internal' | 'config' | 'sync';
declare type FieldInputMethods = 'literal' | 'variable' | 'function' | 'enrichment' | 'freeform';
declare type FieldDisplayMode = 'expanded' | 'normal' | 'collapsed';
export interface InputFieldJSONSchema {
    label: string;
    description: string;
    type: FieldTypeName;
    allowNull?: boolean;
    multiple?: boolean;
    additionalProperties?: boolean;
    default?: FieldValue;
    choices?: Array<string> | Array<{
        value: string | number;
        label: string;
    }>;
    required?: boolean;
    properties?: Record<string, Optional<InputField, 'description'>>;
    format?: 'date' | 'time' | 'date-time' | 'uri' | 'uri-reference' | 'uri-template' | 'email' | 'hostname' | 'ipv4' | 'ipv6' | 'regex' | 'uuid' | 'password' | 'text';
}
export interface InputField extends InputFieldJSONSchema {
    placeholder?: string;
    dynamic?: boolean;
    defaultObjectUI?: 'keyvalue' | 'object' | 'keyvalue:only' | 'object:only' | 'arrayeditor';
    unsafe_hidden?: boolean;
    readOnly?: boolean;
    depends_on?: DependsOnConditions;
    category?: FieldCategory;
    displayMode?: FieldDisplayMode;
    disabledInputMethods?: FieldInputMethods[];
    minimum?: number;
    maximum?: number;
}
interface BaseCondition {
    operator: 'is' | 'is_not';
}
export interface FieldCondition extends BaseCondition {
    fieldKey: string;
    type?: 'field';
    value: Omit<FieldValue, 'Directive'> | Array<Omit<FieldValue, 'Directive'>> | undefined;
}
export interface SyncModeCondition extends BaseCondition {
    type: 'syncMode';
    value: SyncMode;
}
export declare type Condition = FieldCondition | SyncModeCondition;
export interface DependsOnConditions {
    match?: 'any' | 'all';
    conditions: Condition[];
}
export declare type FieldValue = string | number | boolean | object | Directive;
export interface IfDirective {
    '@if': {
        exists?: FieldValue;
        then: FieldValue;
        else?: FieldValue;
    };
}
export interface TemplateDirective {
    '@template': string;
}
export interface PathDirective {
    '@path': string;
}
export interface CaseDirective {
    '@case': {
        operator: string;
        value: FieldValue;
    };
}
export declare type Directive = IfDirective | TemplateDirective | PathDirective | CaseDirective;
export declare type RequestExtension<Settings, Payload = undefined> = (data: ExecuteInput<Settings, Payload>) => RequestOptions;
export interface DeletionPayload {
    userId: ID;
    anonymousId: ID;
}
export declare type Deletion<Settings, Return = any> = (request: RequestClient, data: ExecuteInput<Settings, DeletionPayload>) => MaybePromise<Return>;
export declare const syncModeTypes: readonly ["add", "update", "upsert", "delete"];
export declare type SyncMode = typeof syncModeTypes[number];
export interface SyncModeOption {
    label: string;
    value: SyncMode;
}
export interface SyncModeDefinition {
    default: SyncMode;
    label: string;
    description: string;
    choices: SyncModeOption[];
}
export declare type ActionDestinationSuccessResponseType = {
    status: number;
    sent: JSONLikeObject | string;
    body: JSONLikeObject | string;
};
export declare type ActionDestinationErrorResponseType = {
    status: number;
    errortype: keyof typeof ErrorCodes;
    errormessage: string;
    sent?: JSONLikeObject | string;
    body?: JSONLikeObject | string;
};
export declare type ResultMultiStatusNode = ActionDestinationSuccessResponseType | (ActionDestinationErrorResponseType & {
    errorreporter: MultiStatusErrorReporter;
});
export {};
